/*
 * SearchEngineInterface.java    
 * Version 2.2
 * 2001/10/17
 * Written by Victor Liu See-le <victor@n-gon.com>
 * Copyright GPL
 */

package TheSlowArrowofBeauty;

import java.net.*;
import java.io.*;
import java.util.*;

/*
 * class SearchEngineInterface    
 *
 * Interface to a search engine 
 */
class SearchEngineInterface {

    private Properties props;
    private HtmlReader htmlReader;
    private int numSearchEngines = 0;
    private int numValidSearchEngines = 0;
    private String appletSecurityEvader = null;

    SearchEngineInterface() {
        props = new Properties();
        htmlReader = new HtmlReader();
    }

    public void init() throws IOException {
       loadProperties("search-engines.txt");
                
       for (int i = 1; i < numSearchEngines + 1; i++) {
            String cgiCall = props.getProperty(i + ",cgiCall");
            if (cgiCall == null) {
                continue;
            }
            if (analyzeSearchResults(i, cgiCall)) {
                numValidSearchEngines++;
            }
        }
        if (numValidSearchEngines == 0) {
            throw new IOException("No valid search engines available");
        }
    }
            
    public void loadProperties(String fileName) throws IOException {
        InputStream in = getClass().getResourceAsStream(fileName);
        props.load(in);
        String s = props.getProperty("numSearchEngines");
        numSearchEngines = Integer.parseInt(s);
    }

/*
    public void storeProperties(String fileName) throws IOException {
        FileOutputStream out = new FileOutputStream(fileName);
        props.store(out, null);   // JDK1.2
    }
*/

    protected boolean analyzeSearchResults(int whichOne, String cgiCall) 
            throws IOException {
        String domain = props.getProperty(whichOne + ",domain");
        if (domain == null) {
            throw new IOException("Failure to find domain while " + 
                                  "analyzing search results");
        }
        SearchEngineAnalyzer analyzer = new SearchEngineAnalyzer(domain);
        if (appletSecurityEvader != null) {
            analyzer.setAppletSecurityEvader(appletSecurityEvader);
        }
        SearchEngineResultHint hint;
        hint = analyzer.lookForSearchResultHints(cgiCall);
        if (hint == null) {
            return false;
        }
        props.put(whichOne + ",resultPre", hint.prefix);
        props.put(whichOne + ",resultPost", hint.suffix);
        spill("Analyzed search engine " + domain + ": ");
        spill("  prefix=[" + hint.prefix + 
              "] suffix=[" + hint.suffix + "]");
        if (hint.prefix.length() == 0 || hint.suffix.length() == 0) {
            return false;
        } else {
            props.put(whichOne + ",valid", "true");
            return true;
        }
    }
    
    public URL getSearchResult(String keywords) throws IOException {

        int whichEngine = getValidSearchEngine();
        if (whichEngine == -1) { // Error
            throw new IOException("Can't find a valid search engine!");
        }
        
        String cgiCall = props.getProperty(whichEngine + ",cgiCall");
        String advancePage = props.getProperty(whichEngine + ",advancePage");
        String preURL = props.getProperty(whichEngine + ",resultPre");
        String postURL = props.getProperty(whichEngine + ",resultPost");

        if (cgiCall == null) {
            throw new IOException("No cgi call specified for engine #" + 
                    whichEngine);
        }

        if (preURL == null || postURL == null) {
            throw new IOException("No hints found for engine #" + 
                    whichEngine);
        }

        int start = 0;
        if (advancePage != null) {
            start = (int) (Math.random() * 10) * 10;
            advancePage += start;
        }
        
        String queryResults = getSearchEngineResults(cgiCall + 
                                                     keywords + 
                                                     advancePage);
        if (queryResults == null) {
            return null;
        }
        int p1 = queryResults.indexOf(preURL);
        int p2 = queryResults.indexOf(postURL, p1);
        if (p1 == -1 && p2 == -1) {
            if (Debug.DEBUGGING_ON) {
                System.err.print("No results found. ");
                System.err.println("Trying first page");
            }
            // Try again with no advance page
            queryResults = getSearchEngineResults(cgiCall + keywords);
            if (queryResults == null) {
                return null;
            }
            p1 = queryResults.indexOf(preURL);
            p2 = queryResults.indexOf(postURL, p1);
            if (p1 == -1 && p2 == -1) {
                if (Debug.DEBUGGING_ON) {
                    System.err.print("No results found. ");
                    System.err.println("Trying 1st keyword.");
                }
                // Try with only first keyword
                int plus = keywords.indexOf('+');
                if (plus != -1) {
                    keywords = keywords.substring(0, plus);
                }
                queryResults = getSearchEngineResults(cgiCall + keywords);
                if (queryResults == null) {
                    return null;
                }
                p1 = queryResults.indexOf(preURL);
                p2 = queryResults.indexOf(postURL, p1);
                if (p1 == -1 && p2 == -1) {
                    if (Debug.DEBUGGING_ON) {
                        System.err.print("No results found. ");
                        System.err.println("Trying generic keyword.");
                    }
                    // Try with generic keyword
                    queryResults = 
                            getSearchEngineResults(cgiCall + 
                                                   SearchEngineAnalyzer.QUERY);
                    if (queryResults == null) {
                        return null;
                    }
                    p1 = queryResults.indexOf(preURL);
                    p2 = queryResults.indexOf(postURL, p1);
                    if (p1 == -1 && p2 == -1) {
                        // Bad
                        props.put(whichEngine + ",valid", "false");
                        numValidSearchEngines--;
                        throw new IOException("Could not find preURL " + 
                                              preURL + " for engine #" + 
                                              whichEngine);
                    }
                }
            }
        }
        
        // One of them is ok
        int whichResult = (int) (Math.random() * 10) + 1;
        int lenPre = preURL.length();
        int lenPost = postURL.length();
        if (p1 == -1) {
            // Try looking for postURL
            for (int i = 1; i < whichResult; i++) {
                p2 = queryResults.indexOf(postURL, p2 + lenPost);
            }
            p1 = queryResults.lastIndexOf("=", p2);
            if (p1 == -1) {
                throw new IOException("Could not find preURL before postURL");
            }
            p1++;
            char c = queryResults.charAt(p1);
            while ((Character.isWhitespace(c)) ||
                   (c == '"') || 
                   (c == '\'')) {
                p1++;
                c = queryResults.charAt(p1);
            }
        } else {
            for (int i = 1; i < whichResult; i++) {
                p1 = queryResults.indexOf(preURL, p1 + lenPre);
            }
            if (p1 == -1) {
                throw new IOException("Could not find preURL in stream");
            }
            p1 += lenPre;
        }    
    
        // Find end of URL
        p2 = queryResults.indexOf('>', p1);
        char c = queryResults.charAt(p2 - 1);
        while ((Character.isWhitespace(c)) ||
               (c == '"') || 
               (c == '\'')) {
            p2--;
            c = queryResults.charAt(p2 - 1);
        }
            
        try { 
            String s = queryResults.substring(p1, p2);
            if (!s.startsWith("http://")) {
                return HtmlReader.makeFullURL(cgiCall, s);
            } else { 
                return new URL(s); 
            }
        } catch (MalformedURLException e) {
            System.err.println("MalformedURLException: " + e);
            return null;
        }
    }     

    protected int getValidSearchEngine() {
        if (numSearchEngines == 0) {
            return -1;
        }
        int n = numSearchEngines;
        int whichEngine = (int) (Math.random() * n) + 1;
        String isValid = props.getProperty(whichEngine + ",valid");
        if (isValid != null && isValid.equals("true")) {
            return whichEngine;
        } else {
            if (numSearchEngines == 1) {
                return -1;
            } else {
                return getValidSearchEngine();
            }
        }
    }
    
    protected String getSearchEngineResults(String searchCall) 
            throws IOException {
        try {
            URL searchPage = new URL(searchCall);
            if (Debug.DEBUGGING_ON) {
                System.out.print(searchPage);
                System.out.println(" Getting search engine results");
            }
            return htmlReader.read(searchPage);
        } catch (MalformedURLException e) {
            if (Debug.DEBUGGING_ON) {
                System.err.println("MalformedURLException: " + e);
            }
            return null;
        }
    }
    
    public void setAppletSecurityEvader(String s) {
        appletSecurityEvader = s;
        htmlReader.setAppletSecurityEvader(s);
    }
    
    void spill(String s) {
        System.out.println(s);
    }
}
