/*
 * SlowArrowDisplay.java    
 * Version 2.2
 * 2001/10/17
 * Written by Victor Liu See-le <victor@n-gon.com>
 * Copyright GPL
 */

package TheSlowArrowofBeauty;

import java.awt.*;
import java.awt.event.*;
import java.util.*;
import java.net.*;
import java.io.*;
import java.lang.Math;

/**
 * class SlowArrowDisplay
 *
 * Container for displays, keywords, progress indicators
 */
class SlowArrowDisplay extends Panel 
                       implements Runnable,
                                  MouseListener,
                                  MouseMotionListener,
                                  TimerListener {

    private static final int WAITING            = 0;
    private static final int HAVE_URL           = 1;
    private static final int HTML_PROCESSED     = 2;
    private static final int DISPLAY_DONE       = 3;
    private static final int LINK_CLICKED       = 4;
    private static final int STRIKED            = 5;
    private static final int RESTING            = 6;

    private static final int MAX_CONSECUTIVE_GETURL_FAILURES = 12;
    
    SlowArrow parent;
    Color backColor;
    Color textColor;
    Color linkColor;
    Color typeOnColor;
    Color hiliteColor;
    Color clickColor;
    Color strikeColor;
    Color strikeBackColor;
    private SlowArrowProgressIndicator progressIndicator = null;
    private HtmlReader htmlReader;
    private String raw;
    private int linkIndex;
    private String curLink;
    private Vector text;
    private Vector links;
    private volatile Thread thisThread = null;
    private Image img = null;
    private Graphics gc;
    private Font font;
    private FontMetrics fm;
    private int charWidth;
    private int lineHeight;
    private Rectangle displayBounds = null;
    private int curX;
    private int curY;
    private int scrollY;
    private String urlHost;
    private String urlFile;
    private int urlPort;
    private int state;
    private short mouseSensor[][] = null;
    private int mouseSensorW;
    private int mouseSensorH;
    private boolean isScrolling;
    private short hilitedItem;
    private short clickedItem;
    private int httpResponseCode = 0;
    private Log log = null;
    private int linkIndex2Follow;
    private int dragX0;
    private int dragY0;
    private Timer clickTimer;
    private SlowArrowDisplayAnalyzer analyzer = null;
    private int textDisplayed;
    private int sleepDelay;
    private int maxLinks2Scan;
    private int mouseStrikeDistance;
    private int magicNumber;
    private int fontSize;
    private int piBoxSize;
    private int httpStreamBufferLength;
    private int transitionAnimRefresh;
    private int strikeX = -1;
    private int strikeY = -1;
    private int strikeShowDistance;
    private int consecutiveGetURLFailures = 0;
    private boolean interruptFlag = false;
    private String appletSecurityEvader = null;

    SlowArrowDisplay(SlowArrow parent, 
                     SlowArrowProgressIndicator pi) {

        setLayout(null);
        addMouseListener(this);
        addMouseMotionListener(this);
        this.parent = parent;
        this.progressIndicator = pi;
        htmlReader = new HtmlReader();
        text = new Vector(3000);
        links = new Vector(20, 20);

        sleepDelay = 
                parent.conf.getIntegerProperty("sleepDelay");
        maxLinks2Scan = 
                parent.conf.getIntegerProperty("maxLinks2Scan");
        mouseStrikeDistance = 
                parent.conf.getIntegerProperty("mouseStrikeDistance");
        magicNumber = 
                parent.conf.getIntegerProperty("magicNumber");
        fontSize = 
                parent.conf.getIntegerProperty("fontSize");
        piBoxSize = 
                parent.conf.getIntegerProperty("progressIndicatorBoxSize");
        httpStreamBufferLength = 
                parent.conf.getIntegerProperty("httpStreamBufferLength");
        transitionAnimRefresh = 
                parent.conf.getIntegerProperty("transitionAnimRefresh");
        strikeShowDistance = 
                parent.conf.getIntegerProperty("strikeShowDistance");
        strikeBackColor = 
                parent.conf.getColorProperty("strikeBackColor");
        
        font = new Font("Courier", Font.PLAIN, fontSize);
        fm = Toolkit.getDefaultToolkit().getFontMetrics(font);
        charWidth = fm.charWidth(' ');
        lineHeight = fm.getHeight();
        thisThread = new Thread(this);
        hilitedItem = -1;
        clickedItem = -1;
        
        appletSecurityEvader = 
                parent.conf.getStringProperty("appletSecurityEvader");
        htmlReader.setAppletSecurityEvader(appletSecurityEvader);
    }

    public void start() {
        if (img == null) {
            Dimension d = getSize();
            img = createImage(d.width, d.height);
            gc = img.getGraphics();
            gc.setFont(font);
        }
        if (thisThread == null)    {
            thisThread = new Thread(this);
        }
        try {
            thisThread.sleep(1000);
        } catch (InterruptedException e) {}
        state = RESTING;
        thisThread.start();
    }

    public synchronized void stop() {
        if (analyzer != null) {
            analyzer.kill();
            analyzer = null;
        }
        if (clickTimer != null) {
            clickTimer.stop();
            clickTimer = null;
        }
        if (thisThread != null) {
            Thread moribund = thisThread;
            thisThread = null;
            moribund.interrupt();           
        }
        if (log != null) {
            log.close();
        }
    }

    public void run() {
        Thread currentThread = Thread.currentThread();
        while (thisThread == currentThread) {
            switch (state) { 
            case HAVE_URL:
                thisThread.setPriority(Thread.NORM_PRIORITY);
                processHTML();
                break;

            case HTML_PROCESSED:
                if (progressIndicator != null) {
                    progressIndicator.setEnabled(false);
                }
                thisThread.setPriority(Thread.NORM_PRIORITY);
                display();
                break;

            case DISPLAY_DONE:
                if (progressIndicator != null) {
                    progressIndicator.setEnabled(true);
                }
                thisThread.setPriority(Thread.MIN_PRIORITY);
                scanLinks();
                break;

            case LINK_CLICKED:
                thisThread.setPriority(Thread.MAX_PRIORITY - 1);
                if (progressIndicator != null) {
                    progressIndicator.setEnabled(true);
                }
                followLink();
                break;

            case STRIKED:
                thisThread.setPriority(Thread.MAX_PRIORITY - 1);
                if (progressIndicator != null) {
                    progressIndicator.setEnabled(true);
                }
                setURL(getNewURL());
                break;

            case WAITING:
                try {
                    synchronized(this) {
                        while (state == WAITING) {
                            wait();
                        }
                    } 
                } catch (InterruptedException e) {}
                break;
            
            case RESTING:
            default:
                try {
                    Thread.sleep(3000);
                } catch(InterruptedException e) {}
                setURL(getNewURL());
		        state = HAVE_URL;
                break;
            }
        }
    }

    private void setURL(URL u) {
        clear();                         
	    interruptFlag = false;
        while (u == null) {
            u = getNewURL();
        }
        if (Debug.DEBUGGING_ON) {
            spill(u + " Connecting...");
        }

        try {
            raw = htmlReader.read(u);
        } catch (IOException e) {
            if (Debug.DEBUGGING_ON) {
                spill(u + " IOException thrown " + e);
            }
            parent.history.add(u);
            // Try with new URL
            setURL(getNewURL()); 
            return;
        }
        parent.history.add(u);
        urlHost = u.getHost();
        urlPort = u.getPort();
        urlFile = u.getFile();
        state = HAVE_URL;
    }

    private URL getNewURL() {
        URL u = null;
        try {
            u = parent.searchEngines.getSearchResult(chooseKeywords());
        } catch (IOException e) {
            if (Debug.DEBUGGING_ON) {
                spill("IOException getting new URL: " + e);
            }
            if (e.toString().equals("Can't find a valid search engine!")) {
                spill("Stopping.");
                parent.stop();
                return null;
            }
            consecutiveGetURLFailures++;
            if (Debug.DEBUGGING_ON) {
                spill("Consecutive getNewURL() failures: " + 
                      consecutiveGetURLFailures);
            }
            if (consecutiveGetURLFailures >= MAX_CONSECUTIVE_GETURL_FAILURES) {
                spill("Too many consecutiveGetURLFailures!\nStopping.");
                parent.stop();
                return null;
            } else {
                if (Debug.DEBUGGING_ON) {
                    spill("Trying again...");
                }
                return getNewURL();
            }
        }
        consecutiveGetURLFailures = 0;
        return u;
    }

    private String chooseKeywords() {
        Vector keywords = parent.keywordsList.cloneList();
        int numKeywords = keywords.size();
        int sigma = 0;
        for (int i = 0; i < numKeywords; i++) {
            sigma += ((SlowArrowKeyword) keywords.elementAt(i)).weight;
        }

        int n = (int) (Math.random() * magicNumber) + 1;
        //int n = magicNumber;
        Vector v = new Vector(n);
        String keyword;
        for (int i = 0; i < n; i++) {
            do {
                int rnd = (int) (Math.random() * sigma);
                int j, sum = 0;
                for (j = 0; j < numKeywords; j++) {
                    sum += ((SlowArrowKeyword) keywords.elementAt(j)).weight;
                    if (rnd < sum) {
                        break; 
                    }
                }
                keyword = ((SlowArrowKeyword) keywords.elementAt(j)).keyword;
            } while (v.contains(keyword));
            v.addElement(keyword);
        }

        String s = "";
        for (int i = 0; i < n; i++) {
                s += "+" + v.elementAt(i);
        }
        return s.substring(1);
    }

    private void processHTML() {
        int pos;
        int len = raw.length();
        String tag; 
	    int count = 0;

        for (pos = 0; pos < len; pos ++) {
            char c = raw.charAt(pos);

            if (Character.isWhitespace(c)) {
                continue;
            }

            if (c == '<') {
                pos ++;

                // What kind of tag?
                int endTagPos = raw.indexOf('>', pos);
                if (endTagPos == -1) {
                    endTagPos = len;
                }
                int nextSpace = raw.indexOf(' ', pos);
                if (nextSpace > -1 && nextSpace < endTagPos) {
                    tag = raw.substring(pos, nextSpace);
                    tag.trim();
                } else {
                    tag = raw.substring(pos, endTagPos).trim();
                }
                tag = tag.toUpperCase();

                if (tag.compareTo("A") == 0) {
                    String s = raw.substring(pos, endTagPos);
                    String href = getTagAttribute(s, "HREF");
                    if (htmlReader.validLink(href)) {
                        linkIndex = addLink(href);
                    }
                } else if (tag.compareTo("/A") == 0) {
                
                    linkIndex = -1;
                    
                } else if (tag.compareTo("IMG") == 0) {
                
                    String s = raw.substring(pos, endTagPos);
                    String alt = getTagAttribute(s, "ALT");
                    
                    if (alt != "") {
                        StringTokenizer st = new StringTokenizer(alt);
                        while (st.hasMoreTokens()) {
                            text.addElement(new 
                                    SlowArrowDisplayText(st.nextToken(), 
                                                         linkIndex));
                        }
                    }
                    
                } else if (tag.compareTo("BASE") == 0) {
                
                    String s = raw.substring(pos, endTagPos);
                    String href = getTagAttribute(s, "HREF");
                    
                    try {
                        URL u = new URL(href);
                        urlHost = u.getHost();
                        urlPort = u.getPort();
                        urlFile = u.getFile();
                    } catch (MalformedURLException e) {}
                    
                    if (Debug.DEBUGGING_ON) {
                        spill("New Base HREF = " + href + 
                              " -> " + urlHost + 
                              " - " + urlPort + 
                              " - " + urlFile);
                    }
                    
                } else if (tag.compareTo("SCRIPT") == 0) {
                
                    // Ignore everything up to next </SCRIPT>
                    for (;;) {
                    
                        pos = raw.indexOf("</", endTagPos);
                        if (pos == -1) {
                            pos = len;
                            break;
                        }
                        
                        endTagPos = raw.indexOf('>', pos);
                        if (endTagPos == -1) {
                            endTagPos = len - 1;
                        }
                        
                        tag = raw.substring(pos, endTagPos).trim();
                        if (tag.equalsIgnoreCase("SCRIPT")) {
                            break;
                        }
                    }
                    
                } else if (tag.compareTo("STYLE") == 0) {
                
                    // Ignore everything up to next </STYLE>
                    for (;;) {
                    
                        pos = raw.indexOf("</", endTagPos);
                        if (pos == -1) {
                            pos = len;
                            break;
                        }
                        
                        endTagPos = raw.indexOf('>', pos);
                        if (endTagPos == -1) {
                            endTagPos = len - 1;
                        }
                        
                        tag = raw.substring(pos, endTagPos).trim();
                        if (tag.equalsIgnoreCase("STYLE")) {
                            break;
                        }
                    }
                }
                pos = endTagPos;
            } else {
                // Get next word.
                StringBuffer sbuf = new StringBuffer();

                while ((c > '\u0020') && (c != '<')) {
                    sbuf.append(c);
                    if (pos == len - 1) {
                        break;
                    }
                    c = raw.charAt(++pos);
                }
                if (c == '<') {
                    pos --;
                }

                String word = new String(sbuf);
                if (word.indexOf('&') != -1) {
                    word = htmlReader.replaceSpecialChars(word);
                }

                word.trim();
                if (word.length() == 0) {
                    break;
                }

                if (linkIndex != -1) {
                    text.addElement(new 
                            SlowArrowDisplayText(word, linkIndex));
                } else {
                    text.addElement(new SlowArrowDisplayText(word));
                }
            }
            Thread.yield(); 
            if (interruptFlag) {
                return;
            }
        }
        state = HTML_PROCESSED;
    }

    private void display() {

        if (displayBounds == null) {
        
            Dimension d = getSize();
            displayBounds = new Rectangle(2, 
                                          0, 
                                          d.width - 4, 
                                          d.height - piBoxSize - 2);
            mouseSensorW = displayBounds.width / charWidth;
            mouseSensorH = displayBounds.height / lineHeight;
            
            if (mouseSensor == null) {
                mouseSensor = new short[mouseSensorW][mouseSensorH];
                for (int j = 0; j < mouseSensorH; j++) {
                    for (int i = 0; i < mouseSensorW; i++) {
                        mouseSensor[i][j] = -1;
                    }
                }
            }
        }

        displayTransition();

        if (analyzer != null) {
            if (analyzer.isAlive()) {
                analyzer.kill();
            }
            analyzer = null;
        }
        analyzer = new SlowArrowDisplayAnalyzer();
        analyzer.start();

        curX = displayBounds.x;
        curY = displayBounds.y;
        scrollY = 0;
        int numElements = text.size();
        for (int i = 0; i < numElements; i++) {
            // Get next word.
            Object o = text.elementAt(i);

            if (o instanceof SlowArrowDisplayText) {
                SlowArrowDisplayText sadText = (SlowArrowDisplayText) o;
                int len = sadText.text.length();
                int w = charWidth * len;

                // Type on
                gc.setColor(typeOnColor);
                gc.drawString(sadText.text, 
                              curX, 
                              curY - scrollY + lineHeight - 4);
                repaint(curX, curY - scrollY, w, lineHeight);
                try {
                    Thread.sleep(sleepDelay);
                } catch(InterruptedException e) { 
                    if (Debug.DEBUGGING_ON) {
                        spill("Interrupted while displaying");
                    }
                    return;
                }

                // Type off
                gc.setColor(backColor);
                gc.drawString(sadText.text, curX, 
                              curY - scrollY + lineHeight - 4);
                repaint(curX, curY - scrollY, w, lineHeight);
                try {
                    Thread.sleep(sleepDelay);
                } catch(InterruptedException e) { 
                    if (Debug.DEBUGGING_ON) {
                        spill("Interrupted while displaying");
                    }
                    return;
                }

                getNextPosition(w, lineHeight);
                gc.setColor(textColor);
                gc.drawString(sadText.text, curX, 
                              curY - scrollY + lineHeight - 4);

                // Ending punctuation?
                int origW = w;
                int p;
                for (p = len; p > 0; p--) {
                    char c = sadText.text.charAt(p - 1);
                    if (c == ',' || c == '.' || c == '!' || c == ';' || 
                            c == ':' || c == '?' || c == ')' || 
                            c == '\'' || c == '\"' || c == '`' ||
                            c == '}' || c == ']' || c == '>') {
                        continue;
                    } else {
                        break;
                    }
                }
                if (p != len) {
                    sadText.text = sadText.text.substring(0, p);
                    len = p;
                    w = len * charWidth;
                }

                // Beginning punctuation?
                int newX = curX;
                for (p = 0; p < len; p++) {
                    char c = sadText.text.charAt(p);
                    if (c == '(' || c == '\'' || c == '\"' || 
                            c == '{' || c == '[' || c == '<') {
                        continue;
                    } else {
                        break;
                    }
                }
                if (p != 0) {
                    sadText.text = sadText.text.substring(p);
                    len -= p;
                    w = len * charWidth;
                    newX += p * charWidth;
                }

                // Ending 's ?
                if ((len > 2) && 
                        (sadText.text.charAt(len - 2) == '\'') && 
                        (sadText.text.charAt(len - 1) == 's')) {
                    len -= 2;
                    sadText.text = sadText.text.substring(0, len);
                    w = len * charWidth;
                }

                if (sadText.linkIndex != -1) {
                    gc.drawLine(newX, 
                                curY - scrollY + lineHeight - 2,
                                newX + w,curY - scrollY + lineHeight - 2);
                }
                repaint(curX, curY - scrollY, origW, lineHeight);

                // Fill sensor array.
                int arrayX = (curX - displayBounds.x) / charWidth;
                int arrayY = (curY - scrollY - displayBounds.y) / lineHeight;
                for (int k = 0; (k < len) && (k < mouseSensorW); k++) {
                    mouseSensor[arrayX + k][arrayY] = (short) i;
                }

                sadText.x = newX;
                sadText.y = curY;
                sadText.w = w;
                sadText.h = lineHeight;

                textDisplayed = i;

                try {
                    Thread.sleep(sleepDelay);
                } catch(InterruptedException e) { 
                    if (Debug.DEBUGGING_ON) {
                        spill("Interrupted while displaying");
                    }
                    return;
                }

                curX += origW + charWidth;
            }
            if (interruptFlag) {
                return;
            }
        }
        state = DISPLAY_DONE;
    }

    private void displayTransition() {
        Dimension d = getSize();
        int nSteps = 10;
        int rBk = backColor.getRed();
        int gBk = backColor.getGreen();
        int bBk = backColor.getBlue();
        int rHi = hiliteColor.getRed();
        int gHi = hiliteColor.getGreen();
        int bHi = hiliteColor.getBlue();
        int rStep = (rBk - rHi) / nSteps;
        int gStep = (gBk - gHi) / nSteps;
        int bStep = (bBk - bHi) / nSteps;

        gc.setColor(hiliteColor);
        gc.fillRect(0, 0, d.width, d.height);
        repaint();
        try {
            Thread.sleep(transitionAnimRefresh);
        } catch(InterruptedException e) {}

        for (int i = 1; i < nSteps - 1; i++) {
            gc.setColor(new Color(rHi + i * rStep, 
                                  gHi + i * gStep, 
                                  bHi + i * bStep));
            gc.fillRect(0, 0, d.width, d.height);
            repaint();
            try {
                Thread.sleep(transitionAnimRefresh);
            } catch(InterruptedException e) {}
        }

        gc.setColor(backColor);
        gc.fillRect(0, 0, d.width, d.height);
        repaint();
        try {
            Thread.sleep(transitionAnimRefresh);
        } catch(InterruptedException e) {}
    }

    private void getNextPosition(int width, int height) {
        if (!displayBounds.contains(curX + width, curY - scrollY)) {
            curX = displayBounds.x;
            curY += lineHeight;
            while (curY - scrollY > displayBounds.height - lineHeight) {

                isScrolling = true;
                if (hilitedItem != -1) {
                    paintItem(hilitedItem, backColor, textColor);
                }

                for (int i = 0; i < 12; i++) {
                    gc.copyArea(displayBounds.x,
                                displayBounds.y,
                                displayBounds.width,
                                displayBounds.height,
                                0,
                                -lineHeight);
                    gc.setColor(backColor);
                    gc.fillRect(displayBounds.x,
                                displayBounds.y + 
                                        displayBounds.height - 
                                        lineHeight * (i + 1),
                                displayBounds.width,
                                lineHeight);
                    repaint();

                    scrollY += lineHeight;

                    // Shift elements in mouseSensor array one row up.
                    for (int k = 0; k < mouseSensorH - 1; k++) {
                        for (int j = 0; j < mouseSensorW; j++) {
                            mouseSensor[j][k] = mouseSensor[j][k + 1];
                        }
                    }
                    for (int j = 0; j < mouseSensorW; j++) {
                        mouseSensor[j][mouseSensorH - 1] = -1;
                    }

                    try {
                        Thread.sleep(sleepDelay);
                    } catch(InterruptedException e) {}
                }

                isScrolling = false;
            }
        }
    }

    private void scanLinks() {
        int numLinks = links.size();
        int score, highScore = 0;
        URL nextStop = null;
        Vector linkSubset = new Vector(maxLinks2Scan);

        // If too many links, do random subset.
        if (numLinks > maxLinks2Scan) {
            for (int i = 0; i < maxLinks2Scan; i++) {
                String link;
                do {
                    int rnd = (int) (Math.random() * numLinks);
                    link = (String) links.elementAt(rnd);
                } while (linkSubset.contains(link));
                linkSubset.addElement(link);
            }
            numLinks = maxLinks2Scan;
        } else {
            linkSubset = (Vector) links.clone();
        }

        // Read keywords
        Vector keywords = parent.keywordsList.cloneList();
        int numKeywords = keywords.size();

        for (int i = 0; i < numLinks; i++) {
        
            String link = (String) linkSubset.elementAt(i);
            URL u = HtmlReader.makeFullURL(urlHost, urlPort, urlFile, link);
            
            if (Debug.DEBUGGING_ON) {
                spill(u + " Begin link scan");
            }

            // Check history
            if (parent.history.contains(u)) {
                if (Debug.DEBUGGING_ON) {
                    spill(u + " History already contains");
                }
                continue;
            }

            String s = "";
            try {
                s = htmlReader.read(u, 8192);
            } catch (IOException e) {
                continue;
            }
            s = s.trim();
            if (s == "") {
                continue;
            }
            if (Debug.DEBUGGING_ON) {
                spill(u + " Scanning...");
            }
            try {
                Thread.sleep(sleepDelay);
            } catch(InterruptedException e) { 
                if (Debug.DEBUGGING_ON) {
                    spill("Interrupted while scanning");
                }
                return;
            }

            score = 0;
            for (int j = 0; j < numKeywords; j++) {
                SlowArrowKeyword sak = (SlowArrowKeyword) keywords.elementAt(j);
                String keyword = sak.keyword;
                score += tallyKeyword(keyword.toLowerCase(), s, sak.weight);
                score += tallyKeyword(keyword.toUpperCase(), s, sak.weight);
                String mixed = mixedCase(keyword);
                if (mixed != null) {
                    score += tallyKeyword(mixed, s, sak.weight);
                }
            }
            if (Debug.DEBUGGING_ON) {
                spill(u + " score " + score);
            }

            if (score > highScore) {
                nextStop = u;
                highScore = score;
            }
            try {
                Thread.sleep(sleepDelay);
            } catch(InterruptedException e) { 
                if (Debug.DEBUGGING_ON) {
                    spill("Interrupted while scanning");
                }
                return;
            } 

            if (interruptFlag) {
                return;
            }
        }
        if (nextStop != null) {
            setURL(nextStop);
        } else {
            setURL(getNewURL());
        }
    }

    private int tallyKeyword(String key, String search, int inc) {
        int keywordLen = key.length();
        int score = 0;
        int p = 0;
        while ((p = search.indexOf(key, p + keywordLen)) != -1) {
            score += inc;
        }
        return score;
    }

    private void clear() {
        if ((analyzer != null) && (analyzer.isAlive())) {
            analyzer.kill();
        }
        linkIndex = -1;
        text.removeAllElements();
        links.removeAllElements();
        if (mouseSensor != null) {
            for (int j = 0; j < mouseSensorH; j++) {
                for (int i = 0; i < mouseSensorW; i++) {
                    mouseSensor[i][j] = -1;
                }
            }
        }
        isScrolling = false;
        hilitedItem = -1;
        clickedItem = -1;
        textDisplayed = -1;
    }

    private int addLink(String link) {
        if (link.indexOf("mailto:") != -1) {
            return -1; 
        }
        if (link.indexOf("ftp:") != -1) {
            return -1;
        }
        if (link.indexOf("telnet:") != -1) {
            return -1;
        }
        if (link.trim() == "") {
            return -1;
        }
        int p = link.indexOf('#');
        if (p != -1) {
            link = link.substring(0, p);
        }
        if (links.contains(link)) {
            return links.indexOf(link);
        }

        links.addElement(link);
        return (links.size() - 1);
    }

    private String getTagAttribute(String tag, String attribute) {
        String value = "";
        StringTokenizer st = new StringTokenizer(tag, " =");
        int n = st.countTokens();
        for (int i = 0; i < n; i++) {
            String token = st.nextToken();
            if (attribute.equalsIgnoreCase(token)) {
                try { 
                    value = st.nextToken();
                } catch (NoSuchElementException e) {
                    return value;
                }
                if (value.startsWith("\"")) {
                    value = value.substring(1);
                }
                if (value.endsWith("\"")) {
                    value = value.substring(0, value.length() - 1);
                }
                return value;
            }
        }

        return value;
    }

    public void paint(Graphics g) {
        if (img == null) {
            Dimension d = getSize();
            img = createImage(d.width, d.height);
            gc = img.getGraphics();
            gc.setFont(font);
        }
        if (displayBounds != null) {
            g.clipRect(displayBounds.x,  
                       displayBounds.y,  
                       displayBounds.width, 
                       displayBounds.height);
            g.drawImage(img, 0, 0, this);
        }
    }

    public void update(Graphics g) {
        paint(g);
    }

    private void paintItem(int i, Color backColor, Color textColor) {
        paintItem(i, backColor, textColor, false);
    }

    private void paintItem(int i, 
                             Color backColor, 
                             Color textColor, 
                             boolean defer) {
        Object o = text.elementAt(i);
        if (o instanceof SlowArrowDisplayText) {
            SlowArrowDisplayText sadText = (SlowArrowDisplayText) o;
            gc.setColor(backColor);
            gc.fillRect(sadText.x,  
                        sadText.y - scrollY, 
                        sadText.w, 
                        sadText.h);
            gc.setColor(textColor);
            gc.drawString(sadText.text, 
                          sadText.x, 
                          sadText.y - scrollY + lineHeight - 4);
            if (sadText.linkIndex != -1) {
                gc.drawLine(sadText.x, 
                            sadText.y - scrollY + lineHeight - 2, 
                            sadText.x + sadText.w, 
                            sadText.y - scrollY + lineHeight - 2);
            }
            if (defer == false) {
                repaint(sadText.x, 
                        sadText.y - scrollY, 
                        sadText.w, 
                        sadText.h);
            }
        }
    }

    public void mouseMoved(MouseEvent e) {
        if (displayBounds == null) {
            return;
        }
        if (isScrolling) {
            return;
        }
        if (state == LINK_CLICKED || state == STRIKED) {
            return;
        }

        int x = e.getX();
        int y = e.getY(); 
        int arrayX = (x - displayBounds.x) / charWidth;
        int arrayY = (y - displayBounds.y) / lineHeight;
        if ((arrayX >= 0) && 
            (arrayX < mouseSensorW) && 
            (arrayY >= 0) && 
            (arrayY < mouseSensorH)) {
                short i = mouseSensor[arrayX][arrayY];
                if ((hilitedItem != i) && (hilitedItem != -1)) {
                    paintItem(hilitedItem, backColor, textColor);
                }

                hilitedItem = -1;
                if ((i != -1) && (i != clickedItem)) {
                    paintItem(i, hiliteColor, backColor);
                    hilitedItem = i;
                }
        }
    }

    public synchronized void mouseDragged(MouseEvent e) {
        if (displayBounds == null) {
            return;
        }
        if (state == STRIKED) {
            return;
        }
        e.consume();

        int x = e.getX();
        int y = e.getY();
        
        double rad = Math.sqrt(Math.pow((double) (x - dragX0), 2.0) + 
                               Math.pow((double) (y - dragY0), 2.0));

        if (rad > mouseStrikeDistance) {
            rad = (double) mouseStrikeDistance;
        }
        paintStrike(dragX0, dragY0, rad);
        if (rad >= (double) mouseStrikeDistance) {
            state = STRIKED;
	        interruptFlag = true;
            if (Debug.DEBUGGING_ON) {
                spill("Strike");
            }
            return;
        } 
    }

    private void paintStrike(int x0, int y0, double rad) {
        int r = (int) rad;
        int diam = 2 * r - 1;
        int ovalX = x0 - r + 1;
        int ovalY = y0 - r + 1;
        gc.setColor(strikeBackColor);
        gc.fillOval(ovalX, ovalY, diam, diam);
        gc.setColor(strikeColor);
        int root2 = (int) (rad / Math.sqrt(2.0));
        int xX1 = x0 - root2;
        int xX2 = x0 + root2;
        int xY1 = y0 - root2;
        int xY2 = y0 + root2;
        gc.drawLine(xX1, xY1, xX2, xY2);
        gc.drawLine(xX1 + 1, xY1, xX2, xY2 - 1);
        gc.drawLine(xX1, xY1 + 1, xX2 - 1, xY2);
        gc.drawLine(xX1, xY2, xX2, xY1);
        gc.drawLine(xX1, xY2 - 1, xX2 - 1, xY1);
        gc.drawLine(xX1 + 1, xY2, xX2, xY1 + 1);
        repaint(ovalX, ovalY, diam, diam);
    }

    public void mouseClicked(MouseEvent e) {}

    public void mouseEntered(MouseEvent e) {}

    public void mouseExited(MouseEvent e) {
        if (state == LINK_CLICKED) {
            return;
        }

        if (hilitedItem != -1) {
            paintItem(hilitedItem, backColor, textColor);
        }
    }

    public void mousePressed(MouseEvent e) {
        if (displayBounds == null) {
            return;
        } 

        int x = e.getX();
        int y = e.getY(); 

        if (!isScrolling) {
            int arrayX = (x - displayBounds.x) / charWidth;
            int arrayY = (y - displayBounds.y) / lineHeight;
            if ((arrayX >= 0) && 
                    (arrayX < mouseSensorW) && 
                    (arrayY >= 0) && 
                    (arrayY < mouseSensorH)) {
                int i = (int) mouseSensor[arrayX][arrayY];
                if (i != -1) {
                    Object o = text.elementAt(i);
                    if (o instanceof SlowArrowDisplayText) {
                        if (clickTimer != null) {
                            clickTimer.stop();
                            clickTimer = null;
                            paintItem(clickedItem, backColor, textColor);
                        }
                        paintItem(i, clickColor, backColor);
                        clickedItem = (short) i;
                        clickTimer = new Timer(this, 2000, 1);
                        clickTimer.start();
                    }
                }
            }
        }
        dragX0 = x; 
        dragY0 = y;
    }

    public synchronized void mouseReleased(MouseEvent e) {
        if (displayBounds == null) {
            return;
        }
        e.consume();
        if (state == STRIKED) {
            return;
        }

        int x = e.getX();
        int y = e.getY(); 

        int arrayX = (x - displayBounds.x) / charWidth;
        int arrayY = (y - displayBounds.y) / lineHeight;
        if ((arrayX >= 0) && 
                (arrayX < mouseSensorW) && 
                (arrayY >= 0) && 
                (arrayY < mouseSensorH)) {
                
            int i = (int) mouseSensor[arrayX][arrayY];
            if (i == -1) {
                return;
            }

            Object o = text.elementAt(i);
            if (o instanceof SlowArrowDisplayText) {
                SlowArrowDisplayText sadText = (SlowArrowDisplayText) o;
                if (sadText.linkIndex != -1) {
                    gc.setColor(backColor);
                    gc.fillRect(displayBounds.x, 
                                displayBounds.y, 
                                displayBounds.width, 
                                displayBounds.height);
                    paintItem(i, backColor, textColor);
                    repaint();
                    state = LINK_CLICKED;
                    linkIndex2Follow = sadText.linkIndex;
		            interruptFlag = true;
                }
                parent.keywordsList.addKeyword(sadText.text);
                if (Debug.DEBUGGING_ON) {
                    spill("Added keyword " + sadText.text);
                }
                return;
            }
        }
    }
    
    private void followLink() {
        String link = (String) links.elementAt(linkIndex2Follow);
        if (Debug.DEBUGGING_ON) {
            spill("Following link " + link);
        }
        URL u = HtmlReader.makeFullURL(urlHost, urlPort, urlFile, link);
        setURL(u);
    }

    public void timerTriggered() {
        if (clickedItem > -1) {
            paintItem(clickedItem, backColor, textColor);
        }
        clickedItem = -1;
        clickTimer = null;
    }

    private String mixedCase(String s) {
        if (s.length() < 2) {
            return null;
        } else {
            return s.substring(0, 1).toUpperCase() + 
                    s.substring(1).toLowerCase();
        }
    }

    void setColorScheme(int i) {
        backColor = 
                parent.conf.getColorProperty("displayBackColor" + i);
        textColor = 
                parent.conf.getColorProperty("displayTextColor" + i);
        hiliteColor = 
                parent.conf.getColorProperty("displayHiliteColor" + i);
        typeOnColor = 
                parent.conf.getColorProperty("displayTypeOnColor" + i);
        clickColor = 
                parent.conf.getColorProperty("displayTypeOnColor" + i);
        setBackground(backColor);
        strikeColor = backColor;
    }

    private void openLog(String filename) {
        log = new Log(filename);
    }

    private void spill(String s) {
        if (log != null) {
            log.print(s);
        }
        System.out.println(s);
    }

    private class SlowArrowDisplayAnalyzer extends Thread {
        private Hashtable table;
        private int n;
        private int highScore;
        private String leader;
        private Vector keywordMatches;
	    private boolean mustDie = false;

        SlowArrowDisplayAnalyzer() {
            n = text.size();
            table = new Hashtable(1777);
            setPriority(Thread.MIN_PRIORITY);
            highScore = 0;
            keywordMatches = new Vector(parent.maxKeywords);
        }

        public void run() {
            Vector keywords = parent.keywordsList.cloneList();
            int numKeywords = keywords.size();
	        int halfWay = numKeywords / 2;

            try {
                // Count keyword matches.
                for (int i = 0; i < n; i++) {

                    while (i > textDisplayed) {
                        sleep(2 * sleepDelay); // length of time to
			                                   // display one word
                    }

                    if (mustDie || text == null) {
                        throw new InterruptedException("Must die now");
                    }
                    
                    Object o = text.elementAt(i);  
                    
                    if (!(o instanceof SlowArrowDisplayText)) { 
                        continue;
                    }
                    String s = ((SlowArrowDisplayText) o).text.trim();
                    int len = s.length();
                    int j;
                    
                    // Longer than 2 characers
                    if (len < 3) {
                        continue;
                    }
                    
                    // Only alphanumeric
                    for (j = 0; j < len; j++) {
                        if (!Character.isLetterOrDigit(s.charAt(j))) {
                            break;
                        }
                    }
                    if (j < len) {
                        continue;
                    }
		    
                    if ((parent.stopwords != null) && 
                            (parent.stopwords.contains(s.toLowerCase()))) {
                        continue;
                    }
                    
                    // Check if already a keyword.
                    for (j = 0; j < numKeywords; j++) { 
                        // rest twice per loop
                        if (j == 0 || j == halfWay) {
                            sleep(sleepDelay);
                        }                      
                        if (mustDie) {
                            throw new InterruptedException("Must die now");
                        }
                        String keyword = 
                                ((SlowArrowKeyword) 
                                        keywords.elementAt(j)).keyword;
                        if (s.equalsIgnoreCase(keyword)) {
                            keywordMatched(keyword);
                            break;
                        }
                    }
                    if (j < numKeywords - 1) {
                        continue;
                    }
                    
                    Object val = table.get(s);
                    if (val == null) {
                        table.put(s, new Integer(1));
                    } else {
                        int count = ((Integer) val).intValue() + 1;
                        table.put(s, new Integer(count));
                        if (count > highScore) {
                            highScore = count;
                            leader = s;
                        }
                    }
                }
            } catch (InterruptedException e) {
                if (Debug.DEBUGGING_ON) {
                    spill(urlHost + " Interrupted analyzer"); 
                }
            } finally {
                if (Debug.DEBUGGING_ON) {
                    spill(urlHost + " Finally: leader=" + leader + 
                          "->" + highScore + " matches=" + 
                          keywordMatches.size());
                }
                if (mustDie) {
                    return;
                }
                if ((parent != null) && 
                        (parent.keywordsList != null) &&
                        (keywordMatches.size() >= parent.numKeywords2Match)) {
                    parent.keywordsList.addKeyword(leader);
                    if (Debug.DEBUGGING_ON) {
                        spill(urlHost + " Analyzer added keyword " + 
                                 leader + " with score " + highScore);
                    }
                }
            }
        }
        
        private void keywordMatched(String keyword) {
            if (Debug.DEBUGGING_ON) {
                spill(urlHost + " Matched keyword " + keyword);
            }
            if (!keywordMatches.contains(keyword)) {
                keywordMatches.addElement(keyword);
            }
        }

        void kill() {
	        mustDie = true;
            this.stop();
        }
    }

    private class SlowArrowDisplayText extends Object {
        public String text;
        public int linkIndex = -1;
        public int x, y, w, h;
    
        SlowArrowDisplayText(String s) {
            text = s;
        }
    
        SlowArrowDisplayText(String s, int i) {
            this(s);
            linkIndex = i;
        }
    }
}

